(     Title:  Convert among binary floating-point raw,
              memory, and stack formats
       File:  rawfloat.fs
  Test file:  rawfloat-test.fs
   Log file:  mixfloat.log
    Version:  0.9.6
    Revised:  October 16, 2020
     Author:  David N. Williams
    License:  LGPL

We believe any of this code derived from other authors to be either in
the public domain or otherwise compatible with the LGPL.  For the sake
of the LGPL, the rest is
)
\ Copyright (C) 2003, 2005, 2009, 2020  David N. Williams
(
This library is free software; you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or at
your option any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
General Public License for more details.

If you take advantage of the option in the LGPL to put a particular
version of this library under the GPL, the author[s] would regard it as
polite if you would put any direct modifications under the LGPL as well,
and include this paragraph in your license notice.  A direct
modification is one that enhances or extends the library in line with
its original concept, as opposed to developing a distinct application or
library which might use it.

The library does not load unless the floating-point machine storage
format is bit-wise contiguous in memory, all little-endian or all
big-endian, and of the same endianness as machine integers.

The library uses raw32, raw64, raw80, and raw128 formats that give an
application access to the bits of floating-point encodings in a portable
way.  The numbers in their names are the storage widths in bits.  They
occupy sufficiently many contiguous Forth cells on the data stack or in
memory, with the most significant cell on the top of the stack or lower
in memory.  Unused bits are padded by zeroes in the most significant
part of the most significant cell, conforming with the Forth convention
for narrow data on the stack.

When the number of bits per cell versus the number of bits needed for
the native fp representation requires two cells, this is the usual
significance ordering style for double numbers in Forth.  For more than
two cells, it extends that style.

The raw format facilitates multicell operations such as shifting, and
its cell-wise big-endianness makes it effectively endian-invariant for
text input and display.

The conversion between the machine format and the appropriate one of the
raw formats is done with the help of the appropriate specific versions
of F@, and F!

Upon inspection, neither here nor in rawfloat-test.fs do we see words
with mixed data/float inputs, outputs, or intermediate calculations. 
Everything should work whether or not the floating-point stack is
separate from the data stack, and has been checked with both options in
pfe.  See the section "ENVIRONMENTAL CONSTANTS" for how to select an
integrated stack in pfe.

This library refuses to load if cells are neither 32 nor 64 bits, and it
works with either case.

It also refuses to load if chars are not 8 bits, and if 8 bits is not
the address unit.

It eventually refuses to load if the default float format is not 32, 64,
80, or 128 bits.

There is an environmental dependence on lower case.
)

\ *** USER WORDS

\ SIGNATURE
\   RAW-FLOAT

\ UTILITY
\   qfaligned  qfalign  QFCREATE  qfpad

\ DATA BIT CONSTANTS
\   bits/au  bits/cell

\ FLOAT BIT CONSTANTS
\   THEN  LITTLE-ENDIAN?  FLITTLE-ENDIAN?

\ RAW FETCH/STORE/DROP
\   raw32!  raw64!  raw80!  raw128!
\   raw32@  raw64@  raw80@  raw128@
\   raw!  raw@
\   raw-drop

\ RAW/FLOAT CONVERSIONS
\   raw32>f  raw64>f  raw80>f  raw128>f
\   f>raw32  f>raw64  f>raw80  f>raw128
\   raw>f  f>raw

\ RAW CONSTANTS
\   RAW-CONSTANT
\   +max-qnan-raw  +min-qnan-raw  +max-snan-raw  +min-snan-raw  +inf-raw
\   +max-n-raw     +min-n-raw     +max-subn-raw  +min-subn-raw  +0-raw
\   -max-n-raw


\ *** SIGNATURE

: RAW-FLOAT ;


\ *** UTILITY

: qfaligned  ( qfaddr -- qfaddr' )  #127 + [ #127 invert ] literal and ;
: qfalign  ( -- )  here qfaligned here - allot ;
: QFCREATE  ( "name" -- )  CREATE qfalign DOES> ( -- a-addr) qfaligned ;

5 cells CONSTANT /qfpad  \ oversized for 64 bits/cell
QFCREATE qfpad /qfpad allot

[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ; [THEN]

\ *** DATA BIT CONSTANTS
(
This code assumes the default float format to be either totally big- or
little-endian in memory, and that when floats are little-endian, so are
cells.  The tests in rawfloat-test.fs should catch the failure of the
latter assumption.
)
decimal

[UNDEFINED] BITS/AU [IF]
s" ADDRESS-UNIT-BITS" ENVIRONMENT? 0=
[IF] cr .( ***Can't determine ADDRESS-UNIT-BITS.) ABORT [THEN]
CONSTANT BITS/AU
[THEN]

bits/au 1 chars * 8 <> [IF] cr
  .( ***1 CHARS is not 8 bits.) ABORT [THEN]

bits/au 8 <> [IF]
  cr .( ***bits/au is ) bits/au . .( bits, not 8 ) ABORT [THEN]

[UNDEFINED] BITS/CELL [IF]
bits/au 1 cells * CONSTANT BITS/CELL
[THEN]

bits/cell #32 = bits/cell #64 = or 0= [IF]
  cr .( ***bits/cell is ) bits/cell . .( bits, not 32 or 64 ) ABORT [THEN]


\ *** FLOAT BIT CONSTANTS

[UNDEFINED] FLOAT-BIT-CONSTANTS [IF] INCLUDE float-bit-constants.fs [THEN]

LITTLE-ENDIAN?    FLITTLE-ENDIAN? and
LITTLE-ENDIAN? 0= FLITTLE-ENDIAN? 0= and or
0= [IF] cr
  .( ***Data and floats have opposite endianness.) ABORT [THEN]


\ *** RAW FETCH/STORE/DROP

bits/cell #32 = [IF]
[UNDEFINED] 3! [IF]
: 3!  ( c b a addr -- )  \ at addr: a b c
  dup >r ! r@ [ 1 cells ] literal + ! r> [ 2 cells ] literal + ! ; 
[THEN]
[UNDEFINED] 3@ [IF]
: 3@  ( addr -- a b c )
  dup >r [ 2 cells ] literal + @ r@ [ 1 cells ] literal + @ r> @  ; 
[THEN]
[UNDEFINED] 4! [IF]
: 4!  ( d c b a addr -- )  \ at addr: a b c d
  [ 4 cells ] literal bounds ( addr+3cells addr)
  DO i ! [ 1 cells ] literal +LOOP ;
[THEN]
[UNDEFINED] 4@ [IF]
: 4@  ( addr -- a b c d )
  [ 3 cells ] literal over + ( addr addr+3cells)
  DO i @ [ -1 cells ] literal +LOOP ; 
[THEN]
[THEN] \ bits/cell

: raw32@  ( addr -- raw32 )    @ ;
: raw32!  ( raw32 addr -- )    ! ;
: raw64@  ( addr -- raw64 )    [ bits/cell #32 = ] [IF] 2@ [ELSE] @  [THEN] ;
: raw64!  ( raw64 addr -- )    [ bits/cell #32 = ] [IF] 2! [ELSE] !  [THEN] ;
: raw80@  ( addr -- raw80 )    [ bits/cell #32 = ] [IF] 3@ [ELSE] 2@ [THEN] ;
: raw80!  ( raw80 addr -- )    [ bits/cell #32 = ] [IF] 3! [ELSE] 2! [THEN] ;
: raw128@  ( addr -- raw128 )  [ bits/cell #32 = ] [IF] 4@ [ELSE] 2@ [THEN] ;
: raw128!  ( raw128 addr -- )  [ bits/cell #32 = ] [IF] 4! [ELSE] 2! [THEN] ;

: raw@  ( addr -- raw )
  [ bits/float #32 = ]  [IF] raw32@  [THEN]
  [ bits/float #64 = ]  [IF] raw64@  [THEN]
  [ bits/float #80 = ]  [IF] raw80@  [THEN]
  [ bits/float #128 = ] [IF] raw128@ [THEN] ;

: raw!  ( raw addr -- )
  [ bits/float #32 = ]  [IF] raw32! [THEN]
  [ bits/float #64 = ]  [IF] raw64! [THEN]
  [ bits/float #80 = ]  [IF] raw80! [THEN]
  [ bits/float #128 = ] [IF] raw128! [THEN] ;

: raw-drop  ( raw -- )
  [ bits/float #32 = ]  [IF] drop [THEN]

  [ bits/float #64 = ]  [IF] [ bits/cell #32 = ]
    [IF] 2drop [ELSE] drop [THEN] [THEN]

  [ bits/float #80 = ]  [IF] [ bits/cell #32 = ]
    [IF] 2drop drop [ELSE] 2drop [THEN] [THEN]

  [ bits/float #128 = ] [IF] [ bits/cell #32 = ]
    [IF] 2drop 2drop [ELSE] 2drop [THEN] [THEN] ;


\ *** RAW/FLOAT CONVERSIONS

: raw32>f  ( raw32 -- f: r )  qfpad raw32! qfpad sf@ ;

: f>raw32  ( f: r -- s: raw32 )
  [ bits/cell #64 = ] [IF] 0 qfpad ! [THEN]  \ init hi-bits to 0
  qfpad sf! qfpad raw32@ ;

: raw64>f  ( raw64 -- f: r )
  [ LITTLE-ENDIAN? bits/cell #32 = and ] [IF] swap [THEN]
  qfpad raw64! qfpad df@ ;

: f>raw64  ( f: r -- s: raw64 )  qfpad df! qfpad raw64@
  [ LITTLE-ENDIAN? bits/cell #32 = and ] [IF] swap [THEN] ;

[DEFINED] XF@ [IF]  \ assume XF! defined, too
: raw80>f  ( raw80 -- f: r )
  [ LITTLE-ENDIAN? ] [IF]
    [ bits/cell #32 = ] [IF] -rot swap [ELSE] swap [THEN]
  [THEN]
  qfpad raw80! qfpad xf@ ;

: f>raw80  ( f: r -- s: raw80 )
  0 qfpad ! \ init hi-bits to 0
  qfpad xf! qfpad raw80@
  [ LITTLE-ENDIAN? ] [IF]
    [ bits/cell #32 = ] [IF] -rot swap [ELSE] swap [THEN]
  [THEN] ;
[THEN]

[DEFINED] QF@ [IF]  \ assume QF! defined, too
: raw128>f  ( raw128 -- f: r )
  [ LITTLE-ENDIAN? ] [IF]
    [ bits/cell #32 = ] [IF] swap 2swap swap [ELSE] swap [THEN]
  [THEN]
  qfpad raw128! qfpad qf@ ;
  
: f>raw128  ( f: r -- s: raw128 )
  qfpad qf! qfpad raw128@
  [ LITTLE-ENDIAN? ] [IF]
    [ bits/cell #32 = ] [IF] swap 2swap swap [ELSE] swap [THEN]
  [THEN] ;
[THEN]

: raw>f  ( raw.default -- f: r )
  [ bits/float  #32 = ]  [IF] raw32>f  [THEN]
  [ bits/float  #64 = ]  [IF] raw64>f  [THEN]
  [ bits/float  #80 = ]  [IF] raw80>f  [THEN]
  [ bits/float  #128 = ] [IF] raw128>f [THEN] ;

: f>raw  ( f: r -- s: raw.default )
  [ bits/float  #32 = ]  [IF] f>raw32  [THEN]
  [ bits/float  #64 = ]  [IF] f>raw64  [THEN]
  [ bits/float  #80 = ]  [IF] f>raw80  [THEN]
  [ bits/float  #128 = ] [IF] f>raw128 [THEN] ;


 \ *** RAW CONSTANTS
(
The following code is endian invariant.
)

[UNDEFINED] 3CONSTANT [IF]
: 3CONSTANT  ( "name" c b a -- )
\ "name" run:  ( -- a b c )
  CREATE ( a) , ( b) ,  ( c) ,
  DOES> dup >r 2 cells + @ r@ 1 cells + @ r> @  ;
[THEN]

[UNDEFINED] 4CONSTANT [IF]
: 4CONSTANT  ( "name" d c b a -- )
\ "name" run:  ( -- a b c d )
  CREATE ( a) , ( b) , ( c) ,  ( d) ,
  DOES>
  3 cells over + ( addr addr+3cells)
  DO i @ -1 cells +LOOP ;
[THEN]

: RAW-CONSTANT  ( "name" raw -- )
[ bits/cell #32 = ] [IF]
  [ bits/float #32 = ]  [IF] CONSTANT  [THEN]
  [ bits/float #64 = ]  [IF] 2CONSTANT [THEN]
  [ bits/float #80 = ]  [IF] 3CONSTANT [THEN]
  [ bits/float #128 = ] [IF] 4CONSTANT [THEN]
[ELSE]  \ 64 bits/cell
  [ bits/float #32 = ]  [IF] CONSTANT  [THEN]
  [ bits/float #64 = ]  [IF] CONSTANT  [THEN]
  [ bits/float #80 = ]  [IF] 2CONSTANT [THEN]
  [ bits/float #128 = ] [IF] 2CONSTANT [THEN]
[THEN] ;

hex
bits/float #32 = [IF]  \ same for 32|64 bits/cell
7fffffff RAW-CONSTANT +max-qnan-raw
ffffffff RAW-CONSTANT -max-qnan-raw
7fc00000 RAW-CONSTANT +min-qnan-raw
ffc00000 RAW-CONSTANT -min-qnan-raw
7fbfffff RAW-CONSTANT +max-snan-raw
ffbfffff RAW-CONSTANT -max-snan-raw
7f800001 RAW-CONSTANT +min-snan-raw
ff800001 RAW-CONSTANT -min-snan-raw
7f800000 RAW-CONSTANT +inf-raw
ff800000 RAW-CONSTANT -inf-raw
7f7fffff RAW-CONSTANT +max-n-raw
ff7fffff RAW-CONSTANT -max-n-raw
00800000 RAW-CONSTANT +min-n-raw
80800000 RAW-CONSTANT -min-n-raw
007fffff RAW-CONSTANT +max-subn-raw
807fffff RAW-CONSTANT -max-subn-raw
00000001 RAW-CONSTANT +min-subn-raw
80000001 RAW-CONSTANT -min-subn-raw
0        RAW-CONSTANT +0-raw
80000000 RAW-CONSTANT -0-raw [THEN]

bits/cell #32 = [IF]

bits/float #64 = [IF]
ffffffff 7fffffff RAW-CONSTANT +max-qnan-raw
ffffffff ffffffff RAW-CONSTANT -max-qnan-raw
00000000 7ff80000 RAW-CONSTANT +min-qnan-raw
00000000 fff80000 RAW-CONSTANT -min-qnan-raw
ffffffff 7ff7ffff RAW-CONSTANT +max-snan-raw
ffffffff fff7ffff RAW-CONSTANT -max-snan-raw
00000001 7ff00000 RAW-CONSTANT +min-snan-raw
00000001 fff00000 RAW-CONSTANT -min-snan-raw
00000000 7ff00000 RAW-CONSTANT +inf-raw
00000000 fff00000 RAW-CONSTANT -inf-raw
ffffffff 7fefffff RAW-CONSTANT +max-n-raw
ffffffff ffefffff RAW-CONSTANT -max-n-raw
00000000 00100000 RAW-CONSTANT +min-n-raw
00000000 80100000 RAW-CONSTANT -min-n-raw
ffffffff 000fffff RAW-CONSTANT +max-subn-raw
ffffffff 800fffff RAW-CONSTANT -max-subn-raw
00000001 00000000 RAW-CONSTANT +min-subn-raw
00000001 80000000 RAW-CONSTANT -min-subn-raw
0        0        RAW-CONSTANT +0-raw
0        80000000 RAW-CONSTANT -0-raw [THEN]

bits/float #80 = [IF]
ffffffff ffffffff 7fff RAW-CONSTANT +max-qnan-raw
ffffffff ffffffff ffff RAW-CONSTANT -max-qnan-raw
00000000 c0000000 7fff RAW-CONSTANT +min-qnan-raw
00000000 c0000000 ffff RAW-CONSTANT -min-qnan-raw
ffffffff bfffffff 7fff RAW-CONSTANT +max-snan-raw
ffffffff bfffffff ffff RAW-CONSTANT -max-snan-raw
00000001 80000000 7fff RAW-CONSTANT +min-snan-raw
00000001 80000000 ffff RAW-CONSTANT -min-snan-raw
00000000 80000000 7fff RAW-CONSTANT +inf-raw
00000000 80000000 ffff RAW-CONSTANT -inf-raw
ffffffff ffffffff 7ffe RAW-CONSTANT +max-n-raw
ffffffff ffffffff fffe RAW-CONSTANT -max-n-raw
00000000 80000000 0001 RAW-CONSTANT +min-n-raw
00000000 80000000 8001 RAW-CONSTANT -min-n-raw
ffffffff 7fffffff 0000 RAW-CONSTANT +max-subn-raw
ffffffff 7fffffff 8000 RAW-CONSTANT -max-subn-raw
00000001 00000000 0000 RAW-CONSTANT +min-subn-raw
00000001 00000000 8000 RAW-CONSTANT -min-subn-raw
0        0        0    RAW-CONSTANT +0-raw
0        0        8000 RAW-CONSTANT -0-raw [THEN]

bits/float #128 = [IF]
ffffffff ffffffff ffffffff 7fffffff RAW-CONSTANT +max-qnan-raw
ffffffff ffffffff ffffffff ffffffff RAW-CONSTANT -max-qnan-raw
00000000 00000000 00000000 7fff8000 RAW-CONSTANT +min-qnan-raw
00000000 00000000 00000000 ffff8000 RAW-CONSTANT -min-qnan-raw
ffffffff ffffffff ffffffff 7fff7fff RAW-CONSTANT +max-snan-raw
ffffffff ffffffff ffffffff ffff7fff RAW-CONSTANT -max-snan-raw
00000001 00000000 00000000 7fff0000 RAW-CONSTANT +min-snan-raw
00000001 00000000 00000000 ffff0000 RAW-CONSTANT -min-snan-raw
00000000 00000000 00000000 7fff0000 RAW-CONSTANT +inf-raw
00000000 00000000 00000000 ffff0000 RAW-CONSTANT -inf-raw
ffffffff ffffffff ffffffff 7ffeffff RAW-CONSTANT +max-n-raw
ffffffff ffffffff ffffffff fffeffff RAW-CONSTANT -max-n-raw
00000000 00000000 00000000 00010000 RAW-CONSTANT +min-n-raw
00000000 00000000 00000000 80010000 RAW-CONSTANT -min-n-raw
ffffffff ffffffff ffffffff 0000ffff RAW-CONSTANT +max-subn-raw
ffffffff ffffffff ffffffff 8000ffff RAW-CONSTANT -max-subn-raw
00000001 00000000 00000000 00000000 RAW-CONSTANT +min-subn-raw
00000001 00000000 00000000 80000000 RAW-CONSTANT -min-subn-raw
0        0        0        0        RAW-CONSTANT +0-raw
0        0        0        80000000 RAW-CONSTANT -0-raw [THEN]

[ELSE]  \ 64 bits/cell

bits/float #64 = [IF]
7fffffffffffffff RAW-CONSTANT +max-qnan-raw
ffffffffffffffff RAW-CONSTANT -max-qnan-raw
7ff8000000000000 RAW-CONSTANT +min-qnan-raw
fff8000000000000 RAW-CONSTANT -min-qnan-raw
7ff7ffffffffffff RAW-CONSTANT +max-snan-raw
fff7ffffffffffff RAW-CONSTANT -max-snan-raw
7ff0000000000001 RAW-CONSTANT +min-snan-raw
fff0000000000001 RAW-CONSTANT -min-snan-raw
7ff0000000000000 RAW-CONSTANT +inf-raw
fff0000000000000 RAW-CONSTANT -inf-raw
7fefffffffffffff RAW-CONSTANT +max-n-raw
ffefffffffffffff RAW-CONSTANT -max-n-raw
0010000000000000 RAW-CONSTANT +min-n-raw
8010000000000000 RAW-CONSTANT -min-n-raw
000fffffffffffff RAW-CONSTANT +max-subn-raw
800fffffffffffff RAW-CONSTANT -max-subn-raw
0000000000000001 RAW-CONSTANT +min-subn-raw
8000000000000001 RAW-CONSTANT -min-subn-raw
0                RAW-CONSTANT +0-raw
8000000000000000 RAW-CONSTANT -0-raw [THEN]

bits/float #80 = [IF]
ffffffffffffffff 7fff RAW-CONSTANT +max-qnan-raw
ffffffffffffffff ffff RAW-CONSTANT -max-qnan-raw
c000000000000000 7fff RAW-CONSTANT +min-qnan-raw
c000000000000000 ffff RAW-CONSTANT -min-qnan-raw
bfffffffffffffff 7fff RAW-CONSTANT +max-snan-raw
bfffffffffffffff ffff RAW-CONSTANT -max-snan-raw
8000000000000001 7fff RAW-CONSTANT +min-snan-raw
8000000000000001 ffff RAW-CONSTANT -min-snan-raw
8000000000000000 7fff RAW-CONSTANT +inf-raw
8000000000000000 ffff RAW-CONSTANT -inf-raw
ffffffffffffffff 7ffe RAW-CONSTANT +max-n-raw
ffffffffffffffff fffe RAW-CONSTANT -max-n-raw
8000000000000000 0001 RAW-CONSTANT +min-n-raw
8000000000000000 8001 RAW-CONSTANT -min-n-raw
7fffffffffffffff 0000 RAW-CONSTANT +max-subn-raw
7fffffffffffffff 8000 RAW-CONSTANT -max-subn-raw
0000000000000001 0000 RAW-CONSTANT +min-subn-raw
0000000000000001 8000 RAW-CONSTANT -min-subn-raw
0                0    RAW-CONSTANT +0-raw
0                8000 RAW-CONSTANT -0-raw  [THEN]

bits/float #128 = [IF]
ffffffffffffffff 7fffffffffffffff RAW-CONSTANT +max-qnan-raw
ffffffffffffffff ffffffffffffffff RAW-CONSTANT -max-qnan-raw
0000000000000000 7fff800000000000 RAW-CONSTANT +min-qnan-raw
0000000000000000 ffff800000000000 RAW-CONSTANT -min-qnan-raw
ffffffffffffffff 7fff7fffffffffff RAW-CONSTANT +max-snan-raw
ffffffffffffffff ffff7fffffffffff RAW-CONSTANT -max-snan-raw
0000000000000001 7fff000000000000 RAW-CONSTANT +min-snan-raw
0000000000000001 ffff000000000000 RAW-CONSTANT -min-snan-raw
0000000000000000 7fff000000000000 RAW-CONSTANT +inf-raw
0000000000000000 ffff000000000000 RAW-CONSTANT -inf-raw
ffffffffffffffff 7ffeffffffffffff RAW-CONSTANT +max-n-raw
ffffffffffffffff fffeffffffffffff RAW-CONSTANT -max-n-raw
0000000000000000 0001000000000000 RAW-CONSTANT +min-n-raw
0000000000000000 8001000000000000 RAW-CONSTANT -min-n-raw
ffffffffffffffff 0000ffffffffffff RAW-CONSTANT +max-subn-raw
ffffffffffffffff 8000ffffffffffff RAW-CONSTANT -max-subn-raw
0000000000000001 0000000000000000 RAW-CONSTANT +min-subn-raw
0000000000000001 8000000000000000 RAW-CONSTANT -min-subn-raw
0                0                RAW-CONSTANT +0-raw
0                8000000000000000 RAW-CONSTANT -0-raw [THEN]

[THEN] \ bits/cell
decimal
